from werkzeug.utils import import_string as wz_import_string, cached_property
from werkzeug.local import LocalProxy, LocalStack
from flask import (url_for, Markup, json, request, _request_ctx_stack,\
                   has_request_context, abort, current_app)
import imp
import functools
import re
from slugify import slugify
import yaml
import speaklater
import os
from contextlib import contextmanager
import subprocess
import click
import inspect
import logging


def import_string(impstr, attr=None):
    """Imports a string. Can import an attribute of the imported
    class/module using a double colon as a separator
    """
    if inspect.isclass(impstr):
        return impstr
    if "::" in impstr:
        impstr, attr = impstr.split("::")
    imported = wz_import_string(impstr)
    if attr is not None:
        return getobjpath(imported, attr)
    return imported


def getobjpath(obj, path):
    """Returns an item or attribute of the object recursively.
    Item names are specified between brackets, eg: [item].
    Attribute names are prefixed with a dot (the first one is optional), eg: .attr
    Example: getobjpath(obj, "attr1.attr2[item].attr3")
    """
    if not path:
        return obj
    if path.startswith("["):
        item = path[1:path.index("]")]
        return getobjpath(obj[item], path[len(item) + 2:])
    if path.startswith("."):
        path = path[1:]
    if "." in path or "[" in path:
        dot_idx = path.find(".")
        bracket_idx = path.find("[")
        if dot_idx == -1 or bracket_idx < dot_idx:
            idx = bracket_idx
            next_idx = idx
        else:
            idx = dot_idx
            next_idx = idx + 1
        attr = path[:idx]
        return getobjpath(getattr(obj, attr), path[next_idx:])
    return getattr(obj, path)


def find_classes_in_module(module, clstypes):
    """Find classes of clstypes in module
    """
    classes = []
    for item in dir(module):
        item = getattr(module, item)
        try:
            for cls in clstypes:
                if issubclass(item, cls) and item != cls:
                    classes.append(item)
        except:
            pass
    return classes


def remove_yaml_frontmatter(source, return_frontmatter=False):
    """If there's one, remove the YAML front-matter from the source
    """
    if source.startswith("---\n"):
        frontmatter_end = source.find("\n---\n", 4)
        if frontmatter_end == -1:
            frontmatter = source
            source = ""
        else:
            frontmatter = source[0:frontmatter_end]
            source = source[frontmatter_end + 5:]
        if return_frontmatter:
            return (source, frontmatter)
        return source
    if return_frontmatter:
        return (source, None)
    return source


def parse_yaml_frontmatter(source):
    source, frontmatter = remove_yaml_frontmatter(source, True)
    if frontmatter:
        return (yaml.safe_load(frontmatter), source)
    return (None, source)


def populate_obj(obj, attrs):
    """Populates an object's attributes using the provided dict
    """
    for k, v in attrs.iteritems():
        setattr(obj, k, v)


def url_for_static(filename, **kwargs):
    """Shortcut function for url_for('static', filename=filename)
    """
    return url_for('static', filename=filename, **kwargs)


def url_for_same(**overrides):
    return url_for(request.endpoint, **dict(dict(request.args,
        **request.view_args), **overrides))


def wrap_in_markup(f):
    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        return Markup(f(*args, **kwargs))
    return wrapper


def make_kwarg_validator(name, validator_func):
    if not isinstance(name, tuple):
        name = (name,)
    def decorator_gen(**kwargs):
        def decorator(func):
            @functools.wraps(func)
            def wrapper(*args, **kw):
                values = {n: kw.get(n) for n in name}
                if not validator_func(**dict(kwargs, **values)):
                    abort(400)
                return func(*args, **kw)
            return wrapper
        return decorator
    return decorator_gen


def kwarg_validator(name):
    def decorator(validator_func):
        return make_kwarg_validator(name, validator_func)
    return decorator


def inject_app_config(app, config, prefix=None):
    for k, v in config.iteritems():
        if prefix:
            k = "%s%s" % (prefix, k)
        app.config[k.upper()] = v


def extract_unmatched_items(items, allowed_keys, prefix=None, uppercase=False):
    unmatched = {}
    for k, v in items.iteritems():
        if k not in allowed_keys:
            if prefix:
                k = "%s%s" % (prefix, k)
            if uppercase:
                k = k.upper()
            unmatched[k] = v
    return unmatched


def deep_update_dict(a, b):
    for k, v in b.iteritems():
        if k not in a:
            a[k] = v
        elif isinstance(a[k], dict) and isinstance(v, dict):
            deep_update_dict(a[k], v)
        elif isinstance(a[k], list) and isinstance(v, list):
            a[k].extend(v)
        elif isinstance(v, list) and not isinstance(a[k], list):
            a[k] = [a[k]] + v
        else:
            a[k] = v
    return a


class UnknownValue(object):
    pass

unknown_value = UnknownValue()


class AttrDict(dict):
    """Dict which keys are accessible as attributes
    """
    def __getattr__(self, name):
        return self[name]

    def __setattr__(self, name, value):
        self[name] = value

    def __delattr__(self, name):
        del self[name]

    def get_or_raise(self, name, message):
        try:
            return self[name]
        except KeyError:
            raise KeyError(message)

    def for_json(self):
        return dict(self)


class JSONEncoder(json.JSONEncoder):
    """A JSONEncoder which always activates the for_json feature
    """
    def __init__(self, *args, **kwargs):
        kwargs["for_json"] = True
        super(JSONEncoder, self).__init__(*args, **kwargs)

    def default(self, o):
        if isinstance(o, speaklater._LazyString):
            return o.value
        if isinstance(o, set):
            return list(o)
        return json.JSONEncoder.default(self, o)


class ShellError(Exception):
    def __init__(self, returncode, stderr):
        super(ShellError, self).__init__(stderr)
        self.returncode = returncode
        self.stderr = stderr

    def __unicode__(self):
        return unicode(self.stderr)


def shell_exec(args, echo=True, fg="green", **kwargs):
    kwargs["stdout"] = subprocess.PIPE
    kwargs["stderr"] = subprocess.STDOUT
    p = subprocess.Popen(args, **kwargs)
    out, _ = p.communicate()
    if p.returncode > 0:
        raise ShellError(p.returncode, out)
    if echo:
        click.secho(out, fg=fg)
    return out


def get_remote_addr():
    if current_app.debug and "__remoteaddr" in request.values:
        return request.values["__remoteaddr"]
    return request.remote_addr



def match_domains(domain, allowed_domains):
    for allowed_domain in allowed_domains:
        if allowed_domain.startswith('^'):
            if re.search(allowed_domain, domain, re.I):
                return True
        elif allowed_domain.lower() == domain.lower():
            return True
    return False


def match_email_domain(email, allowed_domains):
    if not email or not '@' in email:
        return False
    _, domain = email.split('@')
    return match_domains(domain, allowed_domains)


def remove_all_handlers_from_logger(logger):
    for handler in logger.handlers:
        logger.removeHandler(handler)
    return logger


class ExcludeWerkzeugLogFilter(logging.Filter):
    def filter(self, record):
        return record.name != 'werkzeug'


def join_url_rule(rule1, rule2):
    if not rule1:
        return rule2
    if not rule2:
        return rule1
    return (rule1.rstrip('/') + '/' + rule2.lstrip('/')).rstrip('/')



from flask import has_request_context, _request_ctx_stack, flash, request, current_app
import flask_login
from flask_login.utils import _get_user
from frasco.ext import get_extension_state, has_extension
from frasco.ctx import ContextStack
from frasco.utils import match_email_domain, populate_obj
from frasco.models import db
from frasco.geoip import geolocate_country
from contextlib import contextmanager
from werkzeug.local import LocalProxy
import datetime
from .signals import user_signed_up
from .password import update_password
import logging


# this allows to set a current user without a request context
_no_req_ctx_user_stack = ContextStack()
logger = logging.getLogger('frasco.users')


@contextmanager
def user_login_context(user):
    stack = _no_req_ctx_user_stack
    if has_request_context():
        if not hasattr(_request_ctx_stack.top, 'user_stack'):
            _request_ctx_stack.top.user_stack = ContextStack()
        stack = _request_ctx_stack.top.user_stack
    stack.push(user)
    try:
        yield user
    finally:
        stack.pop()


def get_current_user(user=None):
    if user:
        return user
    if user is False:
        return None
    if not has_request_context():
        return _no_req_ctx_user_stack.top
    user_stack = getattr(_request_ctx_stack.top, 'user_stack', None)
    if user_stack and user_stack.top:
        return user_stack.top
    return _get_user()


def get_current_user_if_logged_in(user=None):
    user = get_current_user(user)
    if not user or not user.is_authenticated:
        return
    return user
    

current_user = LocalProxy(get_current_user)


def is_user_logged_in():
    """Checks if the user is logged in
    """
    return current_user and current_user.is_authenticated


def login_user(user, *args, **kwargs):
    state = get_extension_state('frasco_users')
    provider = kwargs.pop('provider', state.options['default_auth_provider_name'])

    for validator in state.login_validators:
        if not validator(user):
            return False

    if not flask_login.login_user(user, *args, **kwargs):
        return False

    user.last_login_at = datetime.datetime.utcnow()
    user.last_login_from = request.remote_addr
    user.last_login_provider = provider
    if not user.auth_providers:
        user.auth_providers = []
    if provider not in user.auth_providers:
        user.auth_providers.append(provider)

    if state.LoginModel:
        login = state.LoginModel()
        login.user = user
        login.login_at = datetime.datetime.utcnow()
        login.login_from = request.remote_addr
        login.login_user_agent = request.user_agent.string
        login.login_provider = provider
        if has_extension('frasco_geoip'):
            try:
                login.login_country = geolocate_country(use_session_cache=False)
            except:
                pass
        db.session.add(login)

    logger.info('User #%s logged in' % user.id)
    return True
    

def signup_user(email_or_user, password=None, provider=None, flash_messages=True, send_signal=True, **kwargs):
    state = get_extension_state('frasco_users')
    if isinstance(email_or_user, state.Model):
        user = email_or_user
    else:
        user = state.Model()
        user.email = email_or_user
    populate_obj(user, kwargs)

    user.signup_provider = provider or state.options['default_auth_provider_name']
    user.auth_providers = [user.signup_provider]
    if has_request_context():
        user.signup_from = request.remote_addr
        if has_extension('frasco_geoip'):
            try:
                user.signup_country = geolocate_country(use_session_cache=False)
            except:
                pass

    validate_user(user, flash_messages=flash_messages)
    if password:
        update_password(user, password, flash_messages=flash_messages)

    db.session.add(user)
    db.session.flush()

    if state.options["send_welcome_email"]:
        from frasco.mail import send_mail
        template = "users/welcome.txt" if state.options["send_welcome_email"] == True else state.options["send_welcome_email"]
        send_mail(user.email, template, user=user)

    logger.info('New signup as #%s' % user.id)
    if send_signal:
        user_signed_up.send(user=user)
    return user


class UserValidationFailedError(Exception):
    def __init__(self, reason=None):
        super(UserValidationFailedError, self).__init__()
        self.reason = reason


def validate_user(user, ignore_self=True, flash_messages=True, raise_error=True, is_signup=False):
    """Validates a new user object before saving it in the database.
    Checks if a password is present unless must_provide_password is False.
    Checks if the username is unique unless the option username_is_unique is set to False.
    If the email column exists on the user object and the option email_is_unique is set to True,
    also checks if the email is unique.
    """
    state = get_extension_state('frasco_users')
    email = user.email.strip().lower()
    has_username = hasattr(user, 'username')
    username = getattr(user, 'username', None)
    if username:
        username = username.strip()

    if state.user_validators and state.override_builtin_user_validation:
        for validator in state.user_validators:
            if not validator(username, email, is_signup):
                if raise_error:
                    raise UserValidationFailedError()
                return False
        return True

    if has_username:
        if not username and state.options["must_provide_username"]:
            if flash_messages and state.options["must_provide_username_message"]:
                flash(state.options["must_provide_username_message"], "error")
            if raise_error:
                raise UserValidationFailedError("username_missing")
            return False
        if username.lower() in state.options['forbidden_usernames']:
            if flash_messages and state.options["username_taken_message"]:
                flash(state.options["username_taken_message"], "error")
            if raise_error:
                raise UserValidationFailedError("username_forbidden")
            return False
        if len(username) < state.options['min_username_length']:
            if flash_messages and state.options["username_too_short_message"]:
                flash(state.options["username_too_short_message"], "error")
            if raise_error:
                raise UserValidationFailedError("username_too_short")
            return False
        if not state.options['allow_spaces_in_username'] and " " in username:
            if flash_messages and state.options["username_has_spaces_message"]:
                flash(state.options["username_has_spaces_message"], "error")
            if raise_error:
                raise UserValidationFailedError("username_has_spaces")
            return False
        if state.options["username_is_unique"]:
            col = state.Model.username if state.options["username_case_sensitive"] else db.func.lower(state.Model.username)
            uname = username if state.options["username_case_sensitive"] else username.lower()
            q = state.Model.query.filter(col == uname)
            if ignore_self and user.id:
                q = q.filter(state.Model.id != user.id)
            if q.count() > 0:
                if flash_messages and state.options["username_taken_message"]:
                    flash(state.options["username_taken_message"], "error")
                if raise_error:
                    raise UserValidationFailedError("username_taken")
                return False

    if not email and state.options["must_provide_email"]:
        if flash_messages and state.options["must_provide_email_message"]:
            flash(state.options["must_provide_email_message"], "error")
        if raise_error:
            raise UserValidationFailedError("email_missing")
        return False

    if email and state.options["email_is_unique"]:
        q = state.Model.query.filter(state.Model.email == email)
        if ignore_self and user.id:
            q = q.filter(state.Model.id != user.id)
        if q.count() > 0:
            if flash_messages and state.options["email_taken_message"]:
                flash(state.options["email_taken_message"], "error")
            if raise_error:
                raise UserValidationFailedError("email_taken")
            return False

    if email and state.options['email_allowed_domains'] is not None:
        if not match_email_domain(email, state.options['email_allowed_domains']):
            if flash_messages and state.options['email_domain_not_allowed_message']:
                flash(state.options['email_domain_not_allowed_message'], 'error')
            if raise_error:
                raise UserValidationFailedError('email_domain_not_allowed')
            return False

    for validator in state.user_validators:
        if not validator(username, email, is_signup):
            if raise_error:
                raise UserValidationFailedError()
            return False

    return True


def check_rate_limit(ip, remote_addr_prop, date_prop, flash_message=True):
    state = get_extension_state('frasco_users')
    since = datetime.datetime.now() - datetime.timedelta(seconds=state.options['rate_limit_period'])
    count = state.Model.query.filter(getattr(state.Model, remote_addr_prop) == ip, getattr(state.Model, date_prop) >= since).count()
    if count >= state.options['rate_limit_count']:
        if flash_message and state.options["rate_limit_reached_message"]:
            flash(state.options["rate_limit_reached_message"], "error")
        return False
    return True



from flask import current_app, g
from frasco.ext import get_extension_state, has_extension
from frasco.upload.backends import StorageBackend, register_upload_backend
from frasco.upload.utils import save_uploaded_file_temporarly
from frasco.tasks import enqueue_task
import boto
import os


@register_upload_backend
class S3StorageBackend(StorageBackend):
    name = 's3'
    default_options = {'filename_prefix': '',
                       'acl': 'public-read',
                       'async': False,
                       'signed_url': False,
                       's3_urls_ttl': 3600,
                       'set_content_disposition_header_with_filename': True,
                       'charset': None,
                       'use_sig_v4': False,
                       'region_name': None,
                       'connect_params': {},
                       'set_contents_headers': {}}

    def save(self, file, filename, force_sync=False):
        kwargs = dict(filename=filename, content_disposition_filename=file.filename,
            bucket=self.options.get('bucket'), acl=self.options['acl'],
            prefix=self.options['filename_prefix'])

        if not force_sync and self.options.get('async') and has_extension('frasco_tasks'):
            tmpname = save_uploaded_file_temporarly(file, filename)
            enqueue_task(upload_file_to_s3, stream_or_filename=tmpname, mimetype=file.mimetype,
                delete_source=True, **kwargs)
        else:
            upload_file_to_s3(file, **kwargs)

    def url_for(self, filename, **kwargs):
        bucket = self.options.get('bucket')
        if self.options['signed_url']:
            bucket = get_s3_connection().get_bucket(bucket)
            bucket_key = bucket.get_key(filename)
            kwargs.setdefault('expires_in', self.options['s3_urls_ttl'])
            return bucket_key.generate_url(**kwargs)
        return 'https://%s.s3.amazonaws.com/%s' % (bucket, filename)

    def delete(self, filename, force_sync=False):
        if not force_sync and self.options['async'] and has_extension('frasco_tasks'):
            enqueue_task('delete_s3_file', filename=filename)
        else:
            delete_s3_file(filename)


def get_s3_options():
    return get_extension_state('frasco_upload').get_backend('s3').options


def get_s3_connection(use_cached=True):
    if use_cached and "boto_s3_connection" in g:
        return g.boto_s3_connection
    
    options = get_s3_options()
    kwargs = {'aws_access_key_id': options.get('access_key'),
              'aws_secret_access_key': options.get('secret_key')}

    if options.get('use_sig_v4'):
        if not boto.config.get('s3', 'use-sigv4'):
            boto.config.add_section('s3')
            boto.config.set('s3', 'use-sigv4', 'True')
        if options.get('region_name'):
            kwargs['host'] = 's3.%s.amazonaws.com' % options['region_name']
        else:
            kwargs['host'] = 's3.amazonaws.com'

    kwargs.update(options['connect_params'])
    if options.get('region_name'):
        return boto.s3.connect_to_region(options['region_name'], **kwargs)
    conn = boto.connect_s3(**kwargs)
    if use_cached:
        g.boto_s3_connection = conn
    return conn


def get_s3_bucket(bucket=None):
    options = get_s3_options()
    if not bucket:
        assert 'bucket' in options, "Missing bucket option"
        bucket = options['bucket']
    return get_s3_connection().get_bucket(bucket)


def upload_file_to_s3(stream_or_filename, filename, bucket=None, prefix=None,
                      acl=None, mimetype=None, charset=None, delete_source=False,
                      content_disposition_filename=None):
    options = get_s3_options()
    bucket = get_s3_bucket(bucket)
    prefix = prefix or options['filename_prefix']
    bucket_key = bucket.new_key(prefix + filename)
    acl = acl or options['acl']
    headers = {}
    if options['set_content_disposition_header_with_filename']:
        headers['Content-Disposition'] = 'attachment;filename="%s"' % (
            content_disposition_filename or filename)
    if mimetype:
        headers['Content-Type'] = mimetype
    if charset or options['charset']:
        if 'Content-Type' not in headers:
            headers['Content-Type'] = 'binary/octet-stream' # S3 default mimetype
        headers['Content-Type'] += '; charset=%s' % (charset or options['charset'])
    headers.update(options['set_contents_headers'])
    is_filename = isinstance(stream_or_filename, (str, unicode))
    if is_filename:
        bucket_key.set_contents_from_filename(stream_or_filename, headers, policy=acl)
    else:
        bucket_key.set_contents_from_string(stream_or_filename.read(), headers, policy=acl)
    if is_filename and delete_source:
        os.unlink(stream_or_filename)


def delete_s3_file(filename, bucket=None, prefix=None):
    options = get_s3_options()
    bucket = get_s3_bucket(bucket)
    prefix = prefix or options['filename_prefix']
    bucket.delete_key(prefix + filename)



from flask import g, current_app
from frasco.ext import *
from frasco.ctx import ContextStack
from frasco.tasks import enqueue_task
from frasco.templating.extensions import RemoveYamlFrontMatterExtension
from jinja_macro_tags import MacroLoader, MacroRegistry
from flask_mail import Mail, email_dispatched
from jinja2 import ChoiceLoader, FileSystemLoader, PackageLoader
from contextlib import contextmanager
import os
import datetime

try:
    from cpickle import pickle
except ImportError:
    import pickle

from .message import create_message, clickable_links, log_message


bulk_connection_context = ContextStack()


class FrascoMailError(Exception):
    pass


class FrascoMailState(ExtensionState):
    def __init__(self, *args):
        super(FrascoMailState, self).__init__(*args)
        self.template_loaders = []
        layout_loader = PackageLoader(__name__, "templates")
        self.jinja_loader = MacroLoader(ChoiceLoader([ChoiceLoader(self.template_loaders), layout_loader]))
        self.jinja_env = None
        self.connection = None

    def add_template_folder(self, path):
        self.template_loaders.append(FileSystemLoader(path))

    def add_templates_from_package(self, pkg_name, pkg_path="emails"):
        self.template_loaders.append(PackageLoader(pkg_name, pkg_path))


class FrascoMail(Extension):
    """Send emails using SMTP
    """
    name = "frasco_mail"
    state_class = FrascoMailState
    prefix_extra_options = "MAIL_"
    defaults = {"default_layout": "layout.html",
                "default_template_vars": {},
                "inline_css": False,
                "auto_render_missing_content_type": True,
                "log_messages": None, # default is app.testing
                "dumped_messages_folder": "email_logs",
                "localized_emails": None,
                "default_locale": None,
                "markdown_options": {},
                "silent_failures": False,
                "send_async": False}

    def _init_app(self, app, state):
        state.client = Mail(app)
        state.connection = None

        state.add_template_folder(os.path.join(app.root_path, "emails"))
        state.jinja_env = app.jinja_env.overlay(loader=state.jinja_loader)
        state.jinja_env.add_extension(RemoveYamlFrontMatterExtension)
        state.jinja_env.macros = MacroRegistry(state.jinja_env) # the overlay methods does not call the constructor of extensions
        state.jinja_env.macros.register_from_template("layouts/macros.html")
        state.jinja_env.default_layout = state.options["default_layout"]
        state.jinja_env.filters['clickable_links'] = clickable_links

        if (state.options["log_messages"] is not None and state.options["log_messages"]) or \
            (state.options["log_messages"] is None and app.testing):
            email_dispatched.connect(lambda msg, app: log_message(msg, app, state.options['dumped_messages_folder']), weak=False)

        state.locale = None
        if has_extension('frasco_babel', app):
            if state.options['default_locale'] is None:
                state.options['default_locale'] = app.extensions.frasco_babel.options['default_locale']
            if state.options['localized_emails'] is None:
                state.options['localized_emails'] = '{locale}/{filename}'


def send_message(msg):
    state = get_extension_state('frasco_mail')
    try:
        if bulk_connection_context.top:
            bulk_connection_context.top.send(msg)
        else:
            state.client.send(msg)
    except Exception as e:
        if not state.options['silent_failures']:
            raise e
        current_app.log_exception(e)


def send_message_async(msg):
    require_extension('frasco_tasks')
    enqueue_task(send_message, msg=msg)


def send_mail(to, template_filename, *args, **kwargs):
    state = get_extension_state('frasco_mail')
    force_sync = kwargs.pop('_force_sync', False)
    msg = create_message(to, template_filename, *args, **kwargs)
    if msg:
        if has_extension('tasks') and state.options['send_async'] and not force_sync:
            send_message_async(msg)
        else:
            send_message(msg)


@contextmanager
def bulk_mail_connection():
    state = get_extension_state('frasco_mail')
    try:
        connection = state.client.connect()
        # simulate entering a with context
        # (flask-mail does not provide a way to connect otherwise)
        connection.__enter__()
    except Exception as e:
        if not state.options['silent_failures']:
            raise e
        current_app.log_exception(e)
        connection = DummyBulkConnection()
    bulk_connection_context.push(connection)
    try:
        yield connection
    finally:
        bulk_connection_context.top.__exit__(None, None, None)
        bulk_connection_context.pop()


class DummyBulkConnection(object):
    def send(self, *args, **kwargs):
        pass

    def __exit__(self, *args):
        pass



from flask import Response, abort
from .ctx import ContextStack, FlagContextStack
from .utils import AttrDict
import functools


try:
    from marshmallow import Schema as MarshmallowSchema
    marshmallow_available = True
except ImportError:
    marshmallow_available = False


disable_marshaller = FlagContextStack()
marshalling_context_stack = ContextStack()
marshalling_context = marshalling_context_stack.make_proxy()


def marshal(rv, marshaller, func=None, args=None, kwargs=None, **marshaller_kwargs):
    with marshalling_context_stack(AttrDict(func=func, args=args, kwargs=kwargs, marshaller_kwargs=marshaller_kwargs)):
        if marshmallow_available and issubclass(marshaller, MarshmallowSchema):
            schema = marshaller(**marshaller_kwargs)
            return schema.dump(rv).data
        elif hasattr(marshaller, '__marshaller__'):
            marshaller = marshaller.__marshaller__
        return marshaller(rv, **marshaller_kwargs)


def marshal_with(marshaller, filter=None, filter_status_code=403, **marshaller_kwargs):
    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            rv = func(*args, **kwargs)
            if not disable_marshaller.top:
                rv = marshal(rv, marshaller, func=func, args=args, kwargs=kwargs, **marshaller_kwargs)
                if filter:
                    rv = filter(rv)
                    if rv is None and filter_status_code:
                        abort(filter_status_code)
            return rv
        wrapper.marshalled_with = marshaller
        return wrapper
    return decorator


def marshal_many_with(marshaller, filter=None, **marshaller_kwargs):
    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            items = func(*args, **kwargs)
            if not disable_marshaller.top:
                items = [marshal(i, marshaller, func=func, args=args, kwargs=kwargs, **marshaller_kwargs) for i in items]
                if filter:
                    items = [i for i in items if filter(i)]
            return items
        wrapper.marshalled_with = marshaller
        return wrapper
    return decorator


def marshal_dict_with(**mapping):
    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            data = func(*args, **kwargs)
            if disable_marshaller.top:
                return data
            out = {}
            for k, v in data.iteritems():
                if k in mapping:
                    out[k] = marshal(v, mapping[k], func=func, args=args, kwargs=kwargs)
                else:
                    out[k] = v
            return out
        wrapper.marshalled_with = mapping
        return wrapper
    return decorator



import socketio
import os
import urlparse
import uuid
import json
from itsdangerous import URLSafeTimedSerializer, BadSignature


class PresenceRedisManager(socketio.RedisManager):
    def __init__(self, *args, **kwargs):
        self.presence_session_id = kwargs.pop('presence_session_id', str(uuid.uuid4()).split('-')[-1])
        self.presence_key_prefix = "presence%s:" % self.presence_session_id
        super(PresenceRedisManager, self).__init__(*args, **kwargs)

    def enter_room(self, sid, namespace, room):
        super(PresenceRedisManager, self).enter_room(sid, namespace, room)
        if room and room != sid:
            self.redis.sadd("%s%s:%s" % (self.presence_key_prefix, namespace, room), sid)
            self.server.emit('%s:joined' % room, {"sid": sid, "info": self.get_member_info(sid, namespace)},
                room=room, skip_sid=sid)

    def leave_room(self, sid, namespace, room):
        super(PresenceRedisManager, self).leave_room(sid, namespace, room)
        if room and room != sid:
            self.redis.srem("%s%s:%s" % (self.presence_key_prefix, namespace, room), sid)
            self.server.emit('%s:left' % room, sid, room=room, skip_sid=sid)

    def get_room_members(self, namespace, room):
        return self.redis.smembers("%s%s:%s" % (self.presence_key_prefix, namespace, room))

    def set_member_info(self, sid, namespace, info):
        self.redis.set("%s%s@%s" % (self.presence_key_prefix, namespace, sid), json.dumps(info))
        for room in self.get_rooms(sid, namespace):
            if not room or room == sid:
                continue
            self.server.emit('%s:member_updated' % room, {"sid": sid, "info": info}, room=room, skip_sid=sid)

    def get_member_info(self, sid, namespace):
        data = self.redis.get("%s%s@%s" % (self.presence_key_prefix, namespace, sid))
        if data:
            try:
                return json.loads(data)
            except:
                pass
        return {}

    def disconnect(self, sid, namespace):
        super(PresenceRedisManager, self).disconnect(sid, namespace)
        self.redis.delete("%s%s@%s" % (self.presence_key_prefix, namespace, sid))

    def cleanup_presence_keys(self):
        keys = self.redis.keys('%s*' % self.presence_key_prefix)
        pipe = self.redis.pipeline()
        for key in keys:
            pipe.delete(key)
        pipe.execute()


def create_app(redis_url='redis://', channel='socketio', secret=None, token_max_age=None):
    mgr = PresenceRedisManager(redis_url, channel=channel)
    sio = socketio.Server(client_manager=mgr, async_mode='eventlet')
    token_serializer = URLSafeTimedSerializer(secret)
    default_ns = '/'

    @sio.on('connect')
    def connect(sid, env):
        if not secret:
            return
        try:
            qs = urlparse.parse_qs(env['QUERY_STRING'])
            if not 'token' in qs:
                return False
            user_info, allowed_rooms = token_serializer.loads(qs['token'][0], max_age=token_max_age)
            env['allowed_rooms'] = allowed_rooms
            if user_info:
                mgr.set_member_info(sid, default_ns, user_info)
        except BadSignature as e:
            return False

    @sio.on('members')
    def get_room_members(sid, data):
        if not data.get('room') or data['room'] not in mgr.get_rooms(sid, default_ns):
            return []
        return {sid: mgr.get_member_info(sid, default_ns) for sid in mgr.get_room_members(default_ns, data['room'])}

    @sio.on('join')
    def join(sid, data):
        if sio.environ[sid].get('allowed_rooms') and data['room'] not in sio.environ[sid]['allowed_rooms']:
            return False
        sio.enter_room(sid, data['room'])
        return get_room_members(sid, data)

    @sio.on('broadcast')
    def room_broadcast(sid, data):
        sio.emit("%s:%s" % (data['room'], data['event']), data.get('data'), room=data['room'], skip_sid=sid)

    @sio.on('leave')
    def leave(sid, data):
        sio.leave_room(sid, data['room'])

    @sio.on('set')
    def set(sid, data):
        mgr.set_member_info(sid, default_ns, data)

    @sio.on('get')
    def get(sid, data):
        return mgr.get_member_info(data['sid'], default_ns)

    return socketio.Middleware(sio)


def _get_env_var(wsgi_env, name, default=None):
    return wsgi_env.get(name, os.environ.get(name, default))


_wsgi_app = None
def wsgi_app(environ, start_response):
    global _wsgi_app
    if not _wsgi_app:
        _wsgi_app = create_app(_get_env_var(environ, 'SIO_REDIS_URL', 'redis://'),
            _get_env_var(environ, 'SIO_CHANNEL', 'socketio'), _get_env_var(environ, 'SIO_SECRET'))
    return _wsgi_app(environ, start_response)


def cleanup_wsgi_app():
    if _wsgi_app:
        _wsgi_app.engineio_app.manager.cleanup_presence_keys()


def run_server(port=8888, **kwargs):
    from eventlet import wsgi
    import eventlet
    eventlet.sleep()
    eventlet.monkey_patch()
    env = dict([("SIO_%s" % k.upper(), v) for k, v in kwargs.items()])
    wsgi.server(eventlet.listen(('', port)), wsgi_app, environ=env)
    cleanup_wsgi_app()


if __name__ == '__main__':
    import argparse
    argparser = argparse.ArgumentParser(prog='tornadopush',
        description='Start tornadopush server')
    argparser.add_argument('-p', '--port', default=8888, type=int,
        help='Port number')
    argparser.add_argument('-r', '--redis-url', default=os.environ.get('SIO_REDIS_URL', 'redis://'), type=str,
        help='Redis URL')
    argparser.add_argument('-c', '--channel', default=os.environ.get('SIO_CHANNEL', 'socketio'), type=str,
        help='Channel')
    argparser.add_argument('-s', '--secret', default=os.environ.get('SIO_SECRET'), type=str,
        help='Secret')
    args = argparser.parse_args()
    run_server(args.port, redis_url=args.redis_url, channel=args.channel, secret=args.secret)



from flask import json
from apispec import APISpec
from frasco.utils import join_url_rule
import re


__all__ = ('build_swagger_spec',)


def build_swagger_spec(api_version):
    spec = APISpec(title="API %s" % api_version.version,
                    version=api_version.version,
                    openapi_version="3.0.2",
                    basePath=api_version.url_prefix)
    for service in api_version.iter_services():
        paths = {}
        tag = {"name": service.name}
        if hasattr(service, 'description'):
            tag["description"] = service.description
        spec.tag(tag)
        for rule, endpoint, func, options in service.iter_endpoints():
            rule = join_url_rule(service.url_prefix, rule)
            path = paths.setdefault(convert_url_args(rule), {})
            for method in options.get('methods', ['GET']):
                op = build_spec_operation(rule, service.name + '_' + endpoint, func, options)
                op['tags'] = [service.name]
                path[method.lower()] = op
        for path, operations in paths.iteritems():
            spec.path(path=path, operations=operations)

    return spec


def build_spec_operation(rule, endpoint, func, options):
    o = {"operationId": endpoint, "parameters": build_spec_params(rule, endpoint, func, options)}
    if func.__doc__:
        o['description'] = func.__doc__
    return o


def build_spec_params(rule, endpoint, func, options):
    method = options.get('methods', ['GET'])[0]
    params = []
    if hasattr(func, 'request_params'):
        url = convert_url_args(rule)
        for p in reversed(func.request_params):
            for pname in p.names:
                loc = "query"
                if ("{%s}" % pname) in url:
                    loc = "path"
                elif method.upper() in ("POST", "PUT"):
                    loc = "formData"
                o = {"name": pname,
                        "type": convert_type_to_spec(p.type),
                        "required": p.required,
                        "in": loc}
                if p.help:
                    o['description'] = p.help
                params.append(o)
    return params


_url_arg_re = re.compile(r"<([a-z]+:)?([a-z0-9_]+)>")
def convert_url_args(url):
    return _url_arg_re.sub(r"{\2}", url)


def convert_type_to_spec(type):
    if type is int:
        return "integer"
    if type is float:
        return "number"
    if type is bool:
        return "boolean"
    return "string"


def build_swagger_client(spec, spec_url, var_name='API', class_name = 'SwaggerClient'):
    return "function %s(options) { options['spec'] = %s; %s.call(this, '%s', options); } %s.prototype = new %s();" % (
            var_name,
            json.dumps(spec),
            class_name,
            spec_url,
            var_name,
            class_name
        )



from frasco.ext import *
from frasco.utils import import_string
from frasco.models import delayed_tx_calls
from flask_rq2 import RQ
from rq import get_current_job
from .job import FrascoJob
import redis.exceptions
import functools
import logging


logger = logging.getLogger('frasco.tasks')


class FrascoTasksState(ExtensionState):
    def schedule_task(self, pattern, import_name_or_func):
        if isinstance(import_name_or_func, str):
            func = import_string(import_name_or_func)
            import_name = import_name_or_func
        else:
            func = import_name_or_func
            import_name = "%s.%s" % (func.__module__, func.__name__)
        try:
            logger.info("Scheduling task %s at %s" % (import_name, pattern))
            return self.rq.get_scheduler().cron(pattern, func, id="cron-%s" % import_name.replace('.', '-'))
        except redis.exceptions.ConnectionError:
            logger.error("Cannot initialize scheduled tasks as no redis connection is available")


class FrascoTasks(Extension):
    name = 'frasco_tasks'
    state_class = FrascoTasksState
    prefix_extra_options = 'RQ_'

    def _init_app(self, app, state):
        if not app.config.get('RQ_REDIS_URL') and has_extension('frasco_redis', app):
            app.config['RQ_REDIS_URL'] = app.extensions.frasco_redis.options['url']

        app.config.setdefault('RQ_JOB_CLASS', 'frasco.tasks.job.FrascoJob')
        if app.testing:
            app.config.setdefault('RQ_ASYNC', False)
        state.rq = RQ(app)

        for import_name, pattern in state.options.get('schedule', {}).items():
            state.schedule_task(pattern, import_name)


def enqueue_task_now(func, *args, **kwargs):
    return enqueue_now(func, args=args, kwargs=kwargs)


def enqueue_now(func, **options):
    queue_name = options.pop('queue', None)
    return get_extension_state('frasco_tasks').rq.get_queue(queue_name).enqueue_call(func, **options)


def enqueue_task(func, *args, **kwargs):
    return enqueue(func, args=args, kwargs=kwargs)


@delayed_tx_calls.proxy
def enqueue(func, **options):
    return enqueue_now(func, **options)


def get_enqueued_job(id):
    state = get_extension_state('frasco_tasks')
    return FrascoJob.fetch(id, connection=state.rq.connection)


def task(**options):
    def wrapper(func):
        setattr(func, 'enqueue', lambda *a, **kw: enqueue(func, args=a, kwargs=kw, **options))
        setattr(func, 'enqueue_now', lambda *a, **kw: enqueue_now(func, args=a, kwargs=kw, **options))
        return func
    return wrapper



from frasco.ext import *
from frasco.utils import extract_unmatched_items
from frasco.users.user import signup_user
from frasco.users.auth import register_authentification_handler
from frasco.models import db
import ldap
from ldap.filter import escape_filter_chars
from flask.signals import Namespace as SignalNamespace


_signals = SignalNamespace()
ldap_login = _signals.signal('users_ldap_login')
ldap_signup = _signals.signal('users_ldap_signup')


class LdapTrackUUIDUserModelMixin(object):
    ldap_uuid = db.Column(db.String)


class FrascoUsersLDAP(Extension):
    name = "frasco_users_ldap"
    defaults = {"server": None,
                "use_tls": False,
                "bind_dn": None,
                "bind_password": None,
                "user_dn": '',
                "user_filter": "(&(objectClass=inetOrgPerson)(uid=%(user)s))",
                "username_attr": "uid",
                "append_username_domain": None,
                "strip_username_domain": True,
                "email_attr": "mail",
                "additional_attrs": {},
                "group_flags": {},
                "group_dn": '',
                "group_filter": "(&(objectclass=groupOfNames)(cn=%(group)s))",
                "group_member_attr": "member",
                "group_member_uid_user_attr": None,
                "track_uuid": False,
                "track_uuid_attr": "ldap_uuid"}

    def _init_app(self, app, state):
        register_authentification_handler(self.authentify)

    @ext_stateful_method
    def connect(self, state, bind=True):
        conn = ldap.initialize(state.options['server'])
        ldap_opts = extract_unmatched_items(state.options, self.defaults)
        for key, value in ldap_opts:
            conn.set_option(getattr(ldap, 'OPT_%s' % key), value)
        if bind and state.options['bind_dn']:
            conn.simple_bind_s(state.options['bind_dn'].encode('utf-8'),
                state.options['bind_password'].encode('utf-8'))
        if state.options['use_tls']:
            conn.start_tls_s()
        return conn

    def search_objects(self, base_dn, filter, conn=None):
        if not conn:
            conn = self.connect()
        return conn.search_s(base_dn, ldap.SCOPE_SUBTREE, filter)

    @ext_stateful_method
    def search_user(self, state, id, conn=None):
        filter = state.options['user_filter'] % {'user': escape_filter_chars(id)}
        rs = self.search_objects(state.options['user_dn'], filter, conn)
        if rs:
            return rs[0]

    @ext_stateful_method
    def search_group(self, state, id, conn=None):
        filter = state.options['group_filter'] % {'group': escape_filter_chars(id)}
        rs = self.search_objects(state.options['group_dn'], filter, conn)
        if rs:
            return rs[0]

    @ext_stateful_method
    def is_member_of(self, state, group_dn, user_dn, member_attr=None, conn=None, ignore_errors=True):
        if not conn:
            conn = self.connect()
        if not member_attr:
            member_attr = state.options['group_member_attr']
        try:
            return bool(conn.compare_s(group_dn, member_attr, user_dn))
        except ldap.LDAPError as e:
            if not ignore_errors:
                raise
            return False

    @ext_stateful_method
    def authentify(self, state, username, password):
        try:
            conn = self.connect()
            if state.options['append_username_domain'] and "@" not in username:
                username += "@" + state.options['append_username_domain']
            ldap_user = self.search_user(username, conn=conn)
            if ldap_user:
                dn, attrs = ldap_user
                self.connect(bind=False).simple_bind_s(dn, password)
                return self._get_or_create_user_from_ldap(dn, attrs, conn=conn)
        except ldap.LDAPError as e:
            self.get_app().log_exception(e)

    @ext_stateful_method
    def _get_or_create_user_from_ldap(self, state, dn, attrs, conn=None):
        UserModel = self.get_app().extensions.frasco_users.Model

        if state.options['track_uuid']:
            user = UserModel.query.filter(
                getattr(UserModel, state.options['track_uuid_attr']) == attrs[state.options['track_uuid']][0]).first()
        else:
            user = UserModel.query.filter(UserModel.email == attrs[state.options['email_attr']][0].lower()).first()

        if user:
            ldap_login.send(user=user, dn=dn, attrs=attrs, conn=conn)
            return user

        username = attrs[state.options['username_attr']][0]
        if "@" in username and state.options['strip_username_domain']:
            username = username.split('@')[0]

        user = UserModel()
        user.email = attrs[state.options['email_attr']][0]
        user.username = username
        if state.options['track_uuid']:
            setattr(user, state.options['track_uuid_attr'],
                attrs[state.options['track_uuid']][0])
        for target, src in state.options['additional_attrs'].iteritems():
            if src in attrs:
                setattr(user, target, attrs[src][0])

        memberships = {}
        for flag, group_dn in state.options['group_flags'].iteritems():
            member = getattr(user, state.options['group_member_uid_user_attr'])\
                        if state.options['group_member_uid_user_attr'] else dn
            if group_dn not in memberships:
                memberships[group_dn] = self.is_member_of(group_dn, member, conn=conn)
            setattr(user, flag, memberships[group_dn])

        try:
            signup_user(user, provider='ldap')
        except:
            return

        ldap_signup.send(user=user, dn=dn, attrs=attrs, conn=conn)
        return user



from flask import json
from flask.config import Config as FlaskConfig
from frasco.utils import deep_update_dict
import os
import yaml
import errno


class Config(FlaskConfig):
    """Subclass of Flask's Config class to add support to load from YAML file
    """

    def from_json(self, filename, silent=False, deep_update=False):
        filename = os.path.join(self.root_path, filename)

        try:
            with open(filename) as json_file:
                obj = json.loads(json_file.read())
        except IOError as e:
            if silent and e.errno in (errno.ENOENT, errno.EISDIR):
                return False
            e.strerror = 'Unable to load configuration file (%s)' % e.strerror
            raise
        return self.from_mapping(obj, _deep_update=deep_update)

    def from_yaml(self, filename, silent=False, deep_update=False):
        filename = os.path.join(self.root_path, filename)

        try:
            with open(filename) as yaml_file:
                obj = yaml.safe_load(yaml_file.read())
        except IOError as e:
            if silent and e.errno in (errno.ENOENT, errno.EISDIR):
                return False
            e.strerror = 'Unable to load configuration file (%s)' % e.strerror
            raise
        return self.from_mapping(obj, _deep_update=deep_update)

    def from_mapping(self, *mapping, **kwargs):
        mappings = []
        if len(mapping) == 1:
            if hasattr(mapping[0], 'items'):
                mappings.append(mapping[0].items())
            else:
                mappings.append(mapping[0])
        elif len(mapping) > 1:
            raise TypeError(
                'expected at most 1 positional argument, got %d' % len(mapping)
            )
        deep_update = kwargs.pop('_deep_update', False)
        mappings.append(kwargs.items())
        for mapping in mappings:
            if deep_update:
                deep_update_dict(self, dict((k.upper(), v) for (k, v) in mapping))
            else:
                for (key, value) in mapping:
                    self[key.upper()] = value
        return True

    def from_file(self, filename, silent=False, deep_update=False):
        if filename.endswith(".py"):
            self.from_pyfile(filename, silent=silent)
        elif filename.endswith(".js") or filename.endswith(".json"):
            self.from_json(filename, silent=silent, deep_update=deep_update)
        elif filename.endswith(".yml"):
            self.from_yaml(filename, silent=silent, deep_update=deep_update)
        else:
            raise RuntimeError("Unknown config file extension")

    def resolve_includes(self, relative_to=".", key="INCLUDE_FILES"):
        for spec in self.pop(key, []):
            if not isinstance(spec, dict):
                spec = {"filename": spec}
            filename = os.path.join(relative_to, spec["filename"])
            self.from_file(filename, silent=spec.get("silent", False),
                deep_update=spec.get("deep_update", False))


def load_config(app, config_filename='config.yml', env=None, root_path=None):
    if not root_path:
        root_path = app.root_path
    config_path = os.path.join(root_path, config_filename)
    app.config.from_yaml(config_path, silent=True)
    env = env or app.config['ENV']
    filename, ext = os.path.splitext(config_path)
    env_filename = filename + "-" + env + ext
    app.config.from_yaml(env_filename, silent=True, deep_update=True)
    app.config.resolve_includes(root_path)



from flask import after_this_request, _request_ctx_stack
from frasco.ctx import ContextStack, DelayedCallsContext
from contextlib import contextmanager
import functools
from .ext import db
import logging


__all__ = ('transaction', 'as_transaction', 'is_transaction', 'delayed_tx_calls')


_transaction_ctx = ContextStack()
delayed_tx_calls = DelayedCallsContext()
logger = logging.getLogger('frasco.models')


@contextmanager
def transaction():
    if not _transaction_ctx.top:
        logger.debug('BEGIN TRANSACTION')
    _transaction_ctx.push()
    delayed_tx_calls.push()
    try:
        yield
        _transaction_ctx.pop()
        if not _transaction_ctx.top:
            logger.debug('COMMIT TRANSACTION')
            db.session.commit()
        else:
            db.session.flush()
        delayed_tx_calls.pop()
    except:
        _transaction_ctx.pop()
        if not _transaction_ctx.top:
            logger.debug('ROLLBACK TRANSACTION')
            db.session.rollback()
        delayed_tx_calls.pop(drop_calls=True)
        raise


def is_transaction():
    return bool(_transaction_ctx.top)


def as_transaction(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        with transaction():
            return func(*args, **kwargs)
    return wrapper
